from typing import Dict, NamedTuple, TYPE_CHECKING
from pymongo.database import Database
from .lazy_collection import LazyCollection
from pymongo.typings import _Pipeline

if TYPE_CHECKING:
    from .lazy_mongo import LazyMongo


class LazyDatabase(NamedTuple):
    mongo: "LazyMongo"
    database: Database
    default_collection_name: str = None  # type: ignore

    def __getitem__(self, key: str):
        return LazyCollection(
            mongo=self.mongo,
            database=self,
            collection=self.database[key],
        )

    def find_one(
        self,
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.find_one(query, project)

    def find(
        self,
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.find(query, project)

    def insert_one(
        self,
        collection: str = None,  # type: ignore
        document: Dict = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.insert_one(document)

    def update_one(
        self,
        collection: str = None,  # type: ignore
        filter: Dict = None,  # type: ignore
        update: Dict = None,  # type: ignore
        **kwargs,
    ):
        coll = self[collection or self.default_collection_name]

        return coll.update_one(filter, update, **kwargs)

    def update_set_one(
        self,
        collection: str = None,  # type: ignore
        filter: Dict = None,  # type: ignore
        document: Dict = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.update_set_one(filter, document)

    def count(
        self,
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.count(query)

    def distinct(
        self,
        key: str,
        collection: str = None,  # type: ignore
    ):
        coll = self[collection or self.default_collection_name]

        return coll.distinct(key)

    def aggregate(
        self,
        pipeline: _Pipeline,
        collection: str = None,  # type: ignore
        **kwargs,
    ):
        coll = self[collection or self.default_collection_name]

        return coll.aggregate(pipeline, **kwargs)



from typing import Dict, NamedTuple, TYPE_CHECKING
from pymongo.collection import Collection
from pymongo.results import UpdateResult
from pymongo.typings import _Pipeline
from pymongo.errors import DuplicateKeyError
from .update_response import UpdateResponse
from .insert_response import InsertResponse

if TYPE_CHECKING:
    from .lazy_database import LazyDatabase
    from .lazy_mongo import LazyMongo

try:
    from simple_chalk import chalk
except:
    from fun_things.not_chalk import NotChalk

    chalk = NotChalk()


class LazyCollection(NamedTuple):
    mongo: "LazyMongo"
    database: "LazyDatabase"
    collection: Collection

    def find_one(
        self,
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        return self.collection.find_one(query, project)

    def find(
        self,
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        return self.collection.find(query, project)

    def __log_acknowledged(self, ok: bool):
        return chalk.green("✓") if ok else chalk.red("✕")

    def __log_duplicate(
        self,
        operation: str,
        e: DuplicateKeyError,
    ):
        details = ""

        if e.details != None:
            details = e.details["keyValue"].items()
            details = [f"{chalk.yellow(k)}={chalk.yellow(v)}" for k, v in details]
            details = chalk.dim.gray(", ").join(details)

        self.mongo._log_warn(
            operation,
            "{0} {1}".format(
                "Duplicate",
                details,
            ),
        )

    def __log_update(
        self,
        operation: str,
        result: UpdateResult,
    ):
        if result.upserted_id != None:
            self.mongo._log_info(
                operation,
                "{0} {1}".format(
                    str(result.upserted_id),
                    self.__log_acknowledged(result.acknowledged),
                ),
            )
            return

        n = result.modified_count

        if n == 0:
            n = chalk.yellow(n)

        self.mongo._log_info(
            operation,
            "{0} of {1} {2}".format(
                n,
                result.matched_count,
                self.__log_acknowledged(result.acknowledged),
            ),
        )

    def insert_one(
        self,
        document: Dict = None,  # type: ignore
    ):
        OPERATION = "InsertOne"

        try:
            result = self.collection.insert_one(document)

            self.mongo._log_info(
                OPERATION,
                "{0} {1}".format(
                    str(result.inserted_id),
                    self.__log_acknowledged(result.acknowledged),
                ),
            )

            return InsertResponse(
                ok=True,
                result=result,
            )

        except DuplicateKeyError as e:
            self.__log_duplicate(OPERATION, e)

            return InsertResponse(
                ok=False,
                is_duplicate=True,
                error=e,
            )

        except Exception as e:
            self.mongo._log_error(OPERATION, e)

            return InsertResponse(
                ok=False,
                error=e,
            )

    def update_one(
        self,
        filter: Dict = None,  # type: ignore
        update: Dict = None,  # type: ignore
        **kwargs,
    ):
        OPERATION = "UpdateOne"

        try:
            result = self.collection.update_one(
                filter=filter,
                update=update,
                **kwargs,
            )

            self.__log_update(OPERATION, result)

            return UpdateResponse(
                ok=True,
                result=result,
            )

        except DuplicateKeyError as e:
            self.__log_duplicate(OPERATION, e)

            return UpdateResponse(
                ok=False,
                is_duplicate=True,
                error=e,
            )

        except Exception as e:
            self.mongo._log_error(OPERATION, e)

            return UpdateResponse(
                ok=False,
                error=e,
            )

    def update_set_one(
        self,
        filter: Dict = None,  # type: ignore
        document: Dict = None,  # type: ignore
    ):
        """
        Shortcut for `$set`.

        Upsert = `False`
        """
        return self.update_one(
            filter,
            update={
                "$set": document,
            },
            upsert=False,
        )

    def count(
        self,
        query: Dict = None,  # type: ignore
    ):
        return self.collection.count_documents(query)

    def distinct(self, key: str):  # type: ignore
        return self.collection.distinct(key)

    def aggregate(self, pipeline: _Pipeline, **kwargs):
        return self.collection.aggregate(
            pipeline,
            **kwargs,
        )



from typing import Dict, Union
from pymongo import MongoClient
from .lazy_database import LazyDatabase
from pymongo.typings import _Pipeline
import fun_things.logger

try:
    from simple_chalk import chalk
except:
    from fun_things.not_chalk import NotChalk

    chalk = NotChalk()


class LazyMongo:
    def __init__(self):
        self.mongo: MongoClient = None  # type: ignore
        self.default_database: str = None  # type: ignore
        self.default_collection: str = None  # type: ignore
        self.logger = fun_things.logger.new("LazyMongo")

    def _log_info(
        self,
        operation: str,
        message,
    ):
        self.__log(
            self.logger.info,
            operation,
            message,
        )

    def _log_warn(
        self,
        operation: str,
        message,
    ):
        self.__log(
            self.logger.warning,
            operation,
            message,
        )

    def _log_error(
        self,
        operation: str,
        message,
    ):
        self.__log(
            self.logger.error,
            operation,
            message,
        )

    def __log(
        self,
        fn,
        operation: str,
        message: str,
    ):
        operation = chalk.gray.dim(operation)

        if message:
            return fn(
                "<%s> %s",
                operation,
                message,
            )

        return fn(operation)

    def connect(self, uri_or_client: Union[str, MongoClient]):
        if isinstance(uri_or_client, MongoClient):
            self.mongo = uri_or_client
            return self

        try:
            self.mongo = MongoClient(uri_or_client)
        except:
            pass

        return self

    def __getitem__(self, key: str):
        return LazyDatabase(
            mongo=self,
            database=self.mongo[key or self.default_database],
            default_collection_name=self.default_collection,
        )

    def find_one(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.find_one(collection, query, project)

    def find(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
        project: Dict = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.find(collection, query, project)

    def insert_one(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        document: Dict = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.insert_one(collection, document)

    def update_one(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        filter: Dict = None,  # type: ignore
        update: Dict = None,  # type: ignore
        **kwargs,
    ):
        db = self[database or self.default_database]

        return db.update_one(
            collection,
            filter,
            update,
            **kwargs,
        )

    def update_set_one(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        filter: Dict = None,  # type: ignore
        document: Dict = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.update_set_one(
            collection,
            filter,
            document,
        )

    def count(
        self,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        query: Dict = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.count(collection, query)

    def distinct(
        self,
        key: str,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
    ):
        db = self[database or self.default_database]

        return db.distinct(key, collection)

    def aggregate(
        self,
        pipeline: _Pipeline,
        database: str = None,  # type: ignore
        collection: str = None,  # type: ignore
        **kwargs,
    ):
        db = self[database or self.default_database]

        return db.aggregate(pipeline, collection, **kwargs)
